package ttt.model;
import java.util.*;

/** Models a tic-tac-toe game board. */
public class TTTGame extends Observable {
    // constants
    
    /** The tic-tac-toe board's width and height. */
    public static final int BOARD_SIZE = 3;
    
    /** Constants that occupy board squares. */
    public static final Player EMPTY = null;
    
    /*
     * public static final String EMPTY = "", X = "X", O = "O";
     */

    /** Constants for types of events that can occur in the game. */
    public static final String NEW_GAME_EVENT = "NewGame", PLAY_EVENT = "Play",
            RESET_EVENT = "Reset", UNDO_EVENT = "Undo";
    
    // fields
    private Player[][] board;
    private PlayerList players;
    private int gamesCount;
    private Stack<Move> moves;
    
    /** Constructs a new game board. */
    public TTTGame(String[] names) {
        board = new Player[BOARD_SIZE][BOARD_SIZE];
        players = new PlayerList(this, names);
        moves = new Stack<Move>();
        
        reset();
        newGame();
    }
    
    /**
     * Returns true if it is legal for the current player to make a move on
     * square [row, col].
     */
    public boolean canPlay(int row, int col) {
        return 0 <= row && row < BOARD_SIZE && 0 <= col && col < BOARD_SIZE
                && !isGameOver() && board[row][col] == EMPTY;
    }
    
    /** Returns the current Player whose move it is. */
    public Player getCurrentPlayer() {
        return players.getCurrentPlayer();
    }
    
    /** Returns the contents of square [row, col] on the board. */
    public Player getSquare(int row, int col) {
        return board[row][col];
    }
    
    public Iterator<Player> getPlayerIterator() {
        return players.iterator();
    }
    
    /**
     * Returns the number of the player who has currently won the game. If no
     * one has won the game, returns EMPTY.
     */
    public Player getWinner() {
        // check horizontal rows
        for (int row = 0; row < BOARD_SIZE; row++) {
            Player occupant = getOccupant(row, 0, 0, 1);
            if (occupant != EMPTY)
                return occupant;
        }
        
        // check vertical rows
        for (int col = 0; col < BOARD_SIZE; col++) {
            Player occupant = getOccupant(0, col, 1, 0);
            if (occupant != EMPTY)
                return occupant;
        }
        
        // check diagonal top-left - bottom-right
        Player occupant = getOccupant(0, 0, 1, 1);
        if (occupant != EMPTY)
            return occupant;
        
        // check diagonal bottom-left - top-right
        return getOccupant(BOARD_SIZE - 1, 0, -1, 1);
    }
    
    /** Returns true if the game is over; if someone has won the game. */
    public boolean isGameOver() {
        return getWinner() != EMPTY;
    }
    
    /** Starts a new round of Tic-Tac-Toe. */
    public void newGame() {
        for (int r = 0; r < BOARD_SIZE; r++)
            for (int c = 0; c < BOARD_SIZE; c++)
                board[r][c] = EMPTY;
        // currentPlayer = player1;
        gamesCount++;
        
        moves.clear();
        
        setChanged();
        notifyObservers(NEW_GAME_EVENT);
    }
    
    /**
     * Makes a move on the given [row, col] square on the board. If that square
     * is unavailable, the call has no effect.
     */
    public void play(int row, int col) {
        Move move = new Move(row, col);
        move.play();
    }
    
    /** Clears all previous games played and starts over. */
    public void reset() {
        for (int r = 0; r < BOARD_SIZE; r++)
            for (int c = 0; c < BOARD_SIZE; c++)
                board[r][c] = EMPTY;
        
        gamesCount = 0;
        // player1.reset();
        // player2.reset();
        // currentPlayer = player1;
        moves.clear();
        
        setChanged();
        notifyObservers(RESET_EVENT);
    }
    
    public void undoLastMove() {
        Move move = (Move) moves.peek();
        move.undo();
    }
    
    // Returns number of player who has a tic-tac-toe starting at the given
    // [row, col] and progressing in the given direction; or EMPTY if none.
    private Player getOccupant(int row, int col, int dr, int dc) {
        for (int i = 0, r = row, c = col; i < BOARD_SIZE; i++, r += dr, c += dc)
            if (board[row][col] != board[r][c])
                return EMPTY;
        
        return board[row][col];
    }
    
    public class Move {
        private int row, col;
        
        public Move(int row, int col) {
            this.row = row;
            this.col = col;
            moves.add(this);
        }
        
        public void play() {
            if (canPlay(row, col)) {
                Player currentPlayer = getCurrentPlayer();
                board[row][col] = currentPlayer;
                
                if (isGameOver())
                    currentPlayer.win();
                
                setChanged();
                notifyObservers(PLAY_EVENT);
            }
        }
        
        public void undo() {
            board[row][col] = EMPTY;
            moves.pop();
            
            setChanged();
            notifyObservers(UNDO_EVENT);
        }
    }
}
